
/*
 * Copyright (C) 2008 Tudor Marian (tudorm@cs.cornell.edu)
 */

#include "hashmap.h"

/* returns 0 on success, -errno on failure */
int init_hashmap(struct hashmap *map, int length, int bin_limit,
		void * (*alloc_fn)(unsigned long),
		unsigned long (*hash_fn) (void *),
		int (*cmp_fn)(void *, void *), 
		void (*bin_overflow)(struct hash_head *))
{
	int i;
	if ((length <= 0) || (!map) || (!alloc_fn) || (!hash_fn) || (!cmp_fn))
		return -EINVAL;
	
	if ((bin_limit > 0) && (bin_overflow == NULL))
		return -EINVAL;
	
	if (!(map->array = alloc_fn(length * sizeof(struct hash_chain))))
		return -ENOMEM;

	for (i = 0; i < length; i++)
	{
		map->array[i].len = 0;
		INIT_LIST_HEAD(&map->array[i].head);
	}
	
	map->length = length;
	map->bin_limit = bin_limit;
	map->hash = hash_fn;
	map->cmp = cmp_fn;
	map->overflow = bin_overflow;
	return 0;
}

struct hashmap *create_hashmap(int length, int bin_limit,
		void * (*alloc_fn)(unsigned long),
		unsigned long (*hash_fn) (void *),
		int (*cmp_fn)(void *, void *), 
		void (*bin_overflow)(struct hash_head *))
{
	struct hashmap *map;
	int err = 0;
	
	if (!(map = alloc_fn(sizeof(struct hashmap))))
		return NULL;
	
	if ((err = init_hashmap(map, length, bin_limit, alloc_fn, 
			hash_fn, cmp_fn, bin_overflow)) != 0)
		return NULL;
	
	return map;
}

// what happens if the key pointer is NULL?
int contains_key(struct hashmap *map, void *key)
{
	struct hash_head *item;
	struct hash_chain *chain;
	
	if (unlikely (!key) )
		return 0;
	
	chain = &map->array[map->hash(key) % map->length];
	
	list_for_each_entry(item, &chain->head, list)
	{
		if (map->cmp(key, item->key) == 0)
			return 1;
	}
	return 0;
}

struct hash_head *get_item(struct hashmap *map, void *key)
{
	struct hash_head *item = NULL;
	struct hash_chain *chain;

	if (unlikely (!key) )
		return NULL;
	
	chain = &map->array[map->hash(key) % map->length];
	
	list_for_each_entry(item, &chain->head, list)
	{
		if (map->cmp(key, item->key) == 0)
			return item;
	}
	return NULL;
}

struct hash_head *put_item(struct hashmap *map, struct hash_head *new_item)
{
	struct hash_head *tmp, *item, *ret = NULL;
	struct hash_chain *chain;
	
	if (unlikely ((!new_item) || (!new_item->key)) )
		return NULL;
	
	chain = &map->array[map->hash(new_item->key) % map->length];
	
	list_for_each_entry_safe(item, tmp, &chain->head, list)
	{
		if (map->cmp(new_item->key, item->key) == 0)
		{
			list_del(&item->list);
			ret = item;
			chain->len--;
			break;
		}
	}
	
	/* adds the new element at the head of the list LIFO instead of FIFO */
	list_add(&new_item->list, &chain->head);
	chain->len++;
	
	if ((map->bin_limit > 0) && (chain->len > map->bin_limit))
	{
		/* remove the last element (tail) in the list */
		BUG_ON(list_empty(&chain->head));
		item = list_entry(chain->head.prev, struct hash_head, list);
		list_del(&item->list);
		chain->len--;
		map->overflow(item);
	}
	
	return ret;
}

struct hash_head *remove_item(struct hashmap *map, void *key)
{
	struct hash_head *item, *tmp, *ret = NULL;
	struct hash_chain *chain;
	
	if (unlikely (!key) )
		return NULL;
	
	chain = &map->array[map->hash(key) % map->length];
	
	list_for_each_entry_safe(item, tmp, &chain->head, list)
	{
		if (map->cmp(key, item->key) == 0)
		{
			list_del(&item->list);
			chain->len--;
			ret = item;
			break;
		}
	}
	return ret;
}

void clear(struct hashmap *map, void (*free_item)(struct hash_head *))
{
	struct hash_head *item, *tmp;
	struct hash_chain *chain;
	int i;
	
	if (!free_item)
		return;
	
	for (i = 0; i < map->length; i++)
	{
		chain = &map->array[i];
		list_for_each_entry_safe(item, tmp, &chain->head, list)
		{
			list_del(&item->list);
			chain->len--;
			free_item(item);
		}
	}
}

