
/*
 * netslice implementation (Linux kernel module).
 *
 * Copyright (C) 2009 Tudor Marian <tudor.marian@cornell.edu>
 */

/*
 HOWTO get started:
	- load the driver: 
		`sudo insmod netslice_mod.ko'
	- find the major number assigned to the driver
		`grep netslice_dev /proc/devices'
	- create special file if it doesn't exist (assuming major number 251)
		'mknod node c 251 minor' where minor may be 0-to-NRCPUS
	- run user-space program (open & read, write, poll on file descriptor)
*/

#include "util.h"
#include "netslice.h"

/* this flags exists only to check the performance w/o TX queue placement */
#define TX_SELECT_QUEUE
// TODO: double check this is needed
#define EXPLICIT_IP_ROUTE_HARD

#define	KLOG						"NETSLICE"
#define CHR_DEV_NAME				"netslice_dev"
#define COUNTERS_RESET_STRING		"reset"
//#define USE_KMEM_CACHE_SKB_LIST

#define ONE_K	(1<<10)
#define ONE_MEG	(1<<20)
#define ONE_GIG	(1<<30)
#define MIN_WMEM_MAX	(2*ONE_K)
#define MIN_RMEM_MAX	(2*ONE_K)
#define SKB_SKACK_MTU	1500
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 24)
static int hook_chain = NF_IP_FORWARD; //NF_IP_LOCAL_OUT;
#else
static int hook_chain = NF_INET_FORWARD; //NF_INET_LOCAL_OUT;
#endif
module_param(hook_chain, int, 0000);
MODULE_PARM_DESC(hook_chain, "Netfilter hook chain");
static int hook_prio = NF_IP_PRI_FIRST;
module_param(hook_prio, int, 0000);
MODULE_PARM_DESC(hook_prio, "Netfilter hook priority");
static char* no_iface = "eth5";
module_param(no_iface, charp, 0000);
MODULE_PARM_DESC(no_iface, "Interface that should not be interfered with");
static char *procfile = "netslice";
module_param(procfile, charp, 0000);
MODULE_PARM_DESC(procfile, "The name of the /proc/<file>");
static int wmem_max = 128 * ONE_MEG;
//static int wmem_max = 1024 * ONE_MEG;
module_param(wmem_max, int, 0000);
MODULE_PARM_DESC(wmem_max, "Maximum write buffer");
/* due to packet loss w/ cores on second socket */
static int rmem_max = 128 * ONE_MEG;
//static int rmem_max = 1024 * ONE_MEG;
module_param(rmem_max, int, 0000);
MODULE_PARM_DESC(rmem_max, "Maximum read buffer");

void *procfile_page = NULL; /* a page to read / write the procfile */
struct proc_dir_entry *proc_entry = NULL;
struct nf_hook_ops dedup_nf_hook_ops;
struct nf_hook_ops netslice_nf_hook_ops;
/* character device structures */
static dev_t netslice_dev;
static struct cdev netslice_cdev;
/* number of consecutive minor devices for this device */
static int cdevs_minor;
#ifdef USE_KMEM_CACHE_SKB_LIST
struct kmem_cache *skb_list_cache = NULL;
#ifndef KMEM_CACHE
#define KMEM_CACHE(__struct, __flags) kmem_cache_create(#__struct,\
				sizeof(struct __struct), __alignof__(struct __struct),\
				(__flags), NULL)
#endif /* KMEM_CACHE */
#endif


struct netslice_resources
{
	int no_iface, no_iface_ifindex;
};
static struct netslice_resources netslice_res;

struct netslice_stats
{
	unsigned long cnt;
	int noiface, ipfrags, nomatch, nf_hook_dropped, skb_nonlinear, tx_select_q;
	int max_rcv_burst, rcv_trunc, max_snd_burst;
};

struct skb_list_head
{
	spinlock_t lock;
	struct list_head skbs;
	long len; /* length (number of skbs) */
	long bytes_len; /* length in bytes */
	atomic_t pending_bytes_len; /* length in bytes (pending, for snd) */
	long capacity; /* max bytes allowed */
	long dropped;
};

//#define USE_SKB_LIST_CPU_INFO_DEBUG /* may be expensive on some architectures */
struct skb_list
{
	struct list_head list;
	struct sk_buff *skb;
#ifdef USE_SKB_LIST_CPU_INFO_DEBUG
	int cpu;
#endif
};

struct netslice
{
	ssize_t (*read)(struct file *, char __user *, size_t, loff_t *);
	ssize_t (*write)(struct file *, const char __user *, size_t, loff_t *);
	wait_queue_head_t wait; /* tasks wait on this, both rcv and snd */
	struct skb_list_head rcv_queue;
	struct skb_list_head snd_queue;
	void (*pre_tx_csum)(struct sk_buff *);
	/* allows for atomic filter update, since struct netslice_filter is embedded */
	int (*ip_match)(struct netslice *, struct iphdr *);
	struct netslice_filter filter;
	int cpu;
	atomic_t refcnt;
	struct netslice_stats stats;
};
/* the following construct is PER_CPU only for better cache line alignment */
static DEFINE_PER_CPU(struct netslice, per_cpu_netslice);

/*
 * Don't actually neeed it, and I shaved off some 10% workqueue usage. 
 * Remember, ip_route_me_harder expects skb->dst to be set if CONFIG_NET_NS.
 */
static inline int do_route_me_harder(struct sk_buff **pskb)
{
#ifdef EXPLICIT_IP_ROUTE_HARD
	unsigned addr_type = RTN_UNSPEC;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
	return (ip_route_me_harder(pskb, addr_type));
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,28)
	return (ip_route_me_harder((*pskb), addr_type));
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 0, 0)
	// this is nasty
	struct iphdr *iph = ip_hdr(*pskb);
	struct dst_entry *dst;
	struct flowi fl = {};
	fl.nl_u.ip4_u.daddr = iph->saddr;
	ip_route_output_key(&init_net, (struct rtable **) &dst, &fl);
	skb_dst_set(*pskb, dst);
//	return 0;
	return (ip_route_me_harder((*pskb), addr_type));
#else 
	struct iphdr *iph = ip_hdr(*pskb);
	struct rtable *dst;
	struct flowi4 fl = {};
	fl.daddr = iph->saddr;
	dst = ip_route_output_key(&init_net, &fl);
	skb_dst_set(*pskb, (struct dst_entry*)dst);
	return (ip_route_me_harder((*pskb), addr_type));
	
#endif /* LINUX_VERSION_CODE */
#else
	return 0;
#endif /* EXPLICIT_IP_ROUTE_HARD */
}

static int netslice_open(struct inode *inode, struct file *file);
static int netslice_release(struct inode *inode, struct file *file);
static unsigned int netslice_poll(struct file * file, poll_table *wait);
static ssize_t netslice_aio_read(struct kiocb *iocb, const struct iovec *iov, 
	unsigned long nr_segs, loff_t pos);
static ssize_t netslice_aio_write(struct kiocb *iocb, const struct iovec *iov, 
	unsigned long nr_segs, loff_t pos);
static int netslice_ioctl(struct inode *inode, struct file *file, 
	unsigned int cmd, unsigned long arg);
static long unlocked_netslice_ioctl(struct file *file, unsigned int cmd, 
	unsigned long arg);
static ssize_t netslice_readv(struct file *, char __user *, 
	size_t , loff_t *);
static ssize_t netslice_read_default(struct file *, char __user *, 
	size_t, loff_t *);
static ssize_t netslice_writev(struct file *file, const char __user *, 
	size_t, loff_t *);
static ssize_t netslice_write_default(struct file *file, const char __user *, 
	size_t, loff_t *);

static ssize_t netslice_read(struct file *file, char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	return net_slice->read(file, buf, count, ppos);
}

static ssize_t netslice_write(struct file *file, const char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	return net_slice->write(file, buf, count, ppos);
}

/* the file operations, i.e. all character device functions */
static struct file_operations netslice_fops = 
{
		.open = netslice_open,
		.release = netslice_release,
		.read = netslice_read,
		.write = netslice_write,
		.poll = netslice_poll,
		.aio_read = netslice_aio_read,
		.aio_write = netslice_aio_write,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)		
		.ioctl = netslice_ioctl,
#endif
		.unlocked_ioctl = unlocked_netslice_ioctl,
        .owner = THIS_MODULE,
};

static int dummy_ip_match(struct netslice *tap, struct iphdr *iph){ return 0; }
static int default_ip_match(struct netslice *tap, struct iphdr *iph);

static void ip_checksum_skb(struct sk_buff *skb);
static void transport_ip_checksum_skb(struct sk_buff *skb);
static void dummy_pre_tx_csum(struct sk_buff *skb) {}

static void (*pre_tx_csum_fn[])(struct sk_buff *) = 
{
	dummy_pre_tx_csum, ip_checksum_skb, transport_ip_checksum_skb
};

static int get_pre_tx_csum(struct netslice *net_slice)
{
	int i;
	for (i = 0; i < NETSLICE_PRE_TX_CSUM_MAX; i++)
		if (net_slice->pre_tx_csum == pre_tx_csum_fn[i])
			return i;
	return NETSLICE_PRE_TX_CSUM_INVALID;
}

/* TODO: make generic, currently it is hardcoded, and matches hyperthreads */
#define NETSLICE_TWO_CPUS_PER_SLICE_CPU_OFFSET		(num_online_cpus()/2)
static void set_netslice_cpu_mask(struct netslice *net_slice, 
	struct netslice_cpu_mask *mask)
{
	mask->k_peer = 0;
	mask->u_peer = 0;
	
	mask->k_peer |= (1 << net_slice->cpu);
	BUG_ON(mask->k_peer == 0);
	mask->u_peer |= (1 << 
		(net_slice->cpu + NETSLICE_TWO_CPUS_PER_SLICE_CPU_OFFSET));
	BUG_ON(mask->u_peer == 0);
}

/* control the ``device''
	- change the type of checsum performed on the tx path
	- change emulation mode for readv/writev through read/write
	- change filter rules
*/
static long unlocked_netslice_ioctl(struct file *file, unsigned int cmd, 
	unsigned long arg)
{
	struct netslice *net_slice = file->private_data;
	long err = 0;
	int i;
	
	switch (cmd)
	{
		case NETSLICE_PRE_TX_CSUM_GET:
		{
			for (i = 0; i < NETSLICE_PRE_TX_CSUM_MAX; i++)
				if (net_slice->pre_tx_csum == pre_tx_csum_fn[i])
					break;
			
			if (i >= NETSLICE_PRE_TX_CSUM_MAX)
			{
				printk(KLOG ": Incorrect pre_tx_csum function array\n");
				i = NETSLICE_PRE_TX_CSUM_INVALID;
			}
			return put_user(i, (unsigned int __user *) arg);
		}
		case NETSLICE_PRE_TX_CSUM_SET:
		{
			err = get_user(i, (int __user *) arg);
			if (err)
				break;
			
			if (!netslice_pre_tx_csum_valid(i))
				return -EINVAL;
			net_slice->pre_tx_csum = pre_tx_csum_fn[i];
			break;
		}
		case NETSLICE_READV_WRITEV_GET:
		{
			struct netslice_rw_vect rw_vect;
			rw_vect.flags = 0;
			
			if (net_slice->read != netslice_read_default)
				rw_vect.flags |= NETSLICE_READV_DATAGRAM_READ;
				
			if (net_slice->write != netslice_write_default)
				rw_vect.flags |= NETSLICE_WRITEV_DATAGRAM_WRITE;
			
			return copy_to_user((struct netslice_rw_vect *) arg, &rw_vect, 
				sizeof(rw_vect));
		}		
		case NETSLICE_READV_WRITEV_SET:
		{
			struct netslice_rw_vect rw_vect;
			
			err = copy_from_user(&rw_vect, (struct netslice_rw_vect *) arg, 
				sizeof(rw_vect));
			if (err)
				break;
			
			net_slice->read = (rw_vect.flags & NETSLICE_READV_DATAGRAM_READ) ?
				netslice_readv : netslice_read_default;
			
			net_slice->write = (rw_vect.flags & NETSLICE_WRITEV_DATAGRAM_WRITE) ?
				netslice_writev : netslice_write_default;
			break;
		}
		case NETSLICE_FILTER_GET:
		{
			return copy_to_user((struct netslice_filter *) arg, 
				&net_slice->filter, sizeof(struct netslice_filter));
			break;
		}
		case NETSLICE_FILTER_SET:
		{
			struct netslice_filter new_filter;
			int (*ip_match)(struct netslice *, struct iphdr *);
			
			err = copy_from_user(&new_filter, 
				(struct netslice_filter *) arg, sizeof(new_filter));
			
			if (err)
				break;
			
			/* TODO: validate new_filter */
			
			ip_match = net_slice->ip_match;
			net_slice->ip_match = dummy_ip_match;
			
			memcpy(&net_slice->filter, &new_filter, sizeof(new_filter));
			
			net_slice->ip_match = ip_match;
			break;
		}
		case NETSLICE_CPU_MASK_GET:
		{
			struct netslice_cpu_mask ns_cpu_mask;
			
			set_netslice_cpu_mask(net_slice, &ns_cpu_mask);
			
			return copy_to_user((struct netslice_cpu_mask *) arg, 
				&ns_cpu_mask, sizeof(ns_cpu_mask));
			break;
		}
		default:
			return -ENOIOCTLCMD;
	}
	return err;
}

static int netslice_ioctl(struct inode *inode, struct file *file, 
	unsigned int cmd, unsigned long arg)
{
	return unlocked_netslice_ioctl(file, cmd, arg);
}

/* character device open method */
static int netslice_open(struct inode *inode, struct file *file)
{
	struct netslice *net_slice;
	struct cdev *cdev = inode->i_cdev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
	int cindex = iminor(inode);
#else
	int cindex = inode->i_cindex;
#endif
	
	if (!cdev || cdev != &netslice_cdev)
	{
		printk(KLOG ": Inconsistent open!\n");
		return -ENODEV;
	}
	
	if (cindex < 0 || cindex >= num_online_cpus())
	{
		printk(KLOG ": Invalid minor device number: %d, must be in [%d, %d)\n",
			cindex, 0, num_online_cpus());
		return -ENODEV;
	}

	net_slice = &per_cpu(per_cpu_netslice, cindex);
	if (atomic_inc_return(&net_slice->refcnt) > 1)
	{
		printk(KLOG ": Slices are opened exclusively, one per process\n");
		return -EMFILE;
	}
	file->private_data = net_slice;
	printk(KLOG": Opened cdev minor %d, tap at vaddr 0x%p\n", cindex, net_slice);
	
	return 0;
}

/* character device last close method */
static int netslice_release(struct inode *inode, struct file *file)
{
	/* should remove the netlink filter at this time */
	struct netslice *net_slice = file->private_data;
	atomic_dec(&net_slice->refcnt);
	return 0;
}

static ssize_t netslice_readv(struct file *file, char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	struct skb_list_head *rcv_queue = &net_slice->rcv_queue;
	struct skb_list *skb_item, *tmp;
	struct netslice_iov iov_fast[UIO_FASTIOV];
	struct netslice_iov *iov = iov_fast, *free_iov = NULL;
	struct netslice_iov __user *iov_user = (struct netslice_iov __user *) buf;
	ssize_t retval = 0, ret_bytes, fetched_cnt;
	int seg;
	
	LIST_HEAD(ret_skbs);
	DECLARE_WAITQUEUE(wait, current);
	
	if (unlikely(count == 0))
	{
		retval = 0;
		goto out;
	}
	
	if (unlikely(count > UIO_MAXIOV))
	{
		retval = -EINVAL;
		goto out;
	}
	
	if (count > UIO_FASTIOV)
	{
		iov = kmalloc(count * sizeof(*iov), GFP_KERNEL);
		if (iov == NULL)
		{
			retval = -ENOMEM;
			goto out;
		}
		free_iov = iov;
	}
	
	/* I will be writing back the iov_rlen */
	if (unlikely(!access_ok(VERIFY_WRITE, iov_user, count * sizeof(*iov_user))))
	{
		retval = -EFAULT;
		goto out;
	}
	
	if (unlikely(__copy_from_user(iov, iov_user, count * sizeof(*iov_user))))
	{
		retval = -EFAULT;
		goto out;
	}
	
	/* check parameter validity */
	ret_bytes = fetched_cnt = 0;
	for (seg = 0; seg < count; seg++)
	{
		void __user *ubuf = iov[seg].iov_base;
		ssize_t len = (ssize_t) iov[seg].iov_len;
			
		/* invalid or overflow */
		if (unlikely(len < 0 || (ret_bytes + len < ret_bytes)))
		{
			retval = -EINVAL;
			goto out;
		}
		
		if (unlikely(!access_ok(VERIFY_WRITE, ubuf, len)))
		{
			retval = -EINVAL;
			goto out;
		}
		
		ret_bytes += len;
	}
	
	add_wait_queue(&net_slice->wait, &wait);

	int recv_len_diff = 0;
	int recv_byte_diff = 0;
	for (;;)
	{
		__set_current_state(TASK_INTERRUPTIBLE);
		
		spin_lock_bh(&rcv_queue->lock);
		list_for_each(skb_item->list, &rcv_queue->skbs)
		{
			/* can return at most count items */
		  if (fetched_cnt >= count)
				break;
			
			/* unchain the element */
			list_del(&skb_item->list);
			rcv_queue->bytes_len -= skb_item->skb->len;
			recv_len_diff++;

			
			/* chain to the return list */
			list_add_tail(&skb_item->list, &ret_skbs);
			fetched_cnt++;
		}
		spin_unlock_bh(&rcv_queue->lock);
		
		/* update burst stats */
		if (net_slice->stats.max_rcv_burst < fetched_cnt)
			net_slice->stats.max_rcv_burst = fetched_cnt;
		
		if (!list_empty(&ret_skbs))
			break;
		else if (signal_pending(current))
		{
			retval = -EINTR;
			break;
		}
		else if (file->f_flags & O_NONBLOCK)
		{
			retval = -EAGAIN;
			break;
		}
		schedule(); /* perhaps I should use schedule_timeout() */
	}

	rcv_queue->len -= recv_len_diff;
	
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&net_slice->wait, &wait);
	
	if (unlikely(retval < 0))
		goto out;
	
	seg = 0;
	list_for_each_entry_safe(skb_item, tmp, &ret_skbs, list)
	{
		void __user *ubuf = iov[seg].iov_base;
		ssize_t len = (ssize_t) iov[seg].iov_len;
		ssize_t __user *rlen = &iov_user[seg].iov_rlen;
				
		/* ret_bytes = min_t(ssize_t, len, skb_item->skb->len); */
		ret_bytes = skb_item->skb->len;
		if (unlikely(ret_bytes > len))
		{
			ret_bytes = len;
			net_slice->stats.rcv_trunc++;
		}
		if (unlikely(__copy_to_user(ubuf, skb_item->skb->data, ret_bytes)))
		{
			retval = -EFAULT;
			goto out_free_ret_skbs;
		}
		/* write back the iov_rlen */
		if (unlikely(__put_user(ret_bytes, rlen)))
		{
			retval = -EFAULT;
			goto out_free_ret_skbs;
		}
		
		seg++;
	}
	retval = seg;
	
	out_free_ret_skbs:
	list_for_each_entry_safe(skb_item, tmp, &ret_skbs, list)
	{
		list_del(&skb_item->list);
		kfree_skb(skb_item->skb);
		kfree(skb_item);
	}
	
	out:
	if (free_iov)
		kfree(free_iov);
	return retval;
}

static ssize_t netslice_read_default(struct file *file, char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	struct skb_list_head *rcv_queue = &net_slice->rcv_queue;
	struct skb_list *skb_item = NULL;
	ssize_t retval = 0;

	DECLARE_WAITQUEUE(wait, current);

	if (unlikely(count == 0))
	{
		retval = 0;
		goto out;
	}

	/* check parameter validity */
	if (unlikely(!access_ok(VERIFY_WRITE, buf, count)))
	{
		retval = -EINVAL;
		goto out;
	}
	
	add_wait_queue(&net_slice->wait, &wait);
	
	for (;;)
	{
		skb_item = NULL;
		__set_current_state(TASK_INTERRUPTIBLE);
		
		spin_lock_bh(&rcv_queue->lock);
		if (!list_empty(&rcv_queue->skbs))
		{
			skb_item = list_entry(rcv_queue->skbs.next, struct skb_list, list);
			list_del(&skb_item->list);
			
			rcv_queue->bytes_len -= skb_item->skb->len;
			rcv_queue->len--;
		}
		spin_unlock_bh(&rcv_queue->lock);
		
		if (skb_item)
			break;
		else if (signal_pending(current))
		{
			retval = -EINTR;
			break;
		}
		else if (file->f_flags & O_NONBLOCK)
		{
			retval = -EAGAIN;
			break;
		}
		schedule(); /* perhaps I should use schedule_timeout() */
	}
	
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&net_slice->wait, &wait);
	
	if (unlikely(retval < 0))
		goto out;
	
	/* retval = min_t(int, skb_item->skb->len, count); */
	retval = skb_item->skb->len;
	if (unlikely(retval > count))
	{
		retval = count;
		net_slice->stats.rcv_trunc++;
	}
	/* TODO: write it properly, even if we know that (!skb_is_nonlinear) 
	something like skb_copy_datagram_iovec, however, use __copy_to_user */	
	if (unlikely(__copy_to_user(buf, skb_item->skb->data, retval)))
	{
		retval = -EFAULT;
		goto out_free_skb;
	}
	
	
	out_free_skb:
	kfree_skb(skb_item->skb);
	kfree(skb_item);
	
	out:
	return retval;
}

struct netslice_skb_cb
{
	struct netslice *net_slice;
	int len;
};

/* WARNING!!!! beware of the private sk_buff->cb data, 2.6.28 overwrites my 
netslice at address &(skb->cb[0]), 2.6.24 does not, so I picked an offset */
#define CB_OFFSET			(sizeof(long *))
#define NETSLICE_CB(__skb)	((struct netslice_skb_cb *)&((__skb)->cb[CB_OFFSET]))
static void netslice_skb_destructor(struct sk_buff *skb)
{
	struct netslice *net_slice = NETSLICE_CB(skb)->net_slice;
	struct skb_list_head *snd_queue = &net_slice->snd_queue;
	
	atomic_sub(NETSLICE_CB(skb)->len, &snd_queue->pending_bytes_len);
	/* Hmm, I wonder if there's a way to get rid of all the unnecessary wake_up
	calls, without breaking correctness...*/
	wake_up_interruptible(&net_slice->wait);
}

#ifdef TX_SELECT_QUEUE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)

#include <linux/jhash.h>
/* yanked out of net/core/dev.c since it is not exported :( */
static u32 simple_tx_hashrnd;
static int simple_tx_hashrnd_initialized = 0;

static u16 simple_tx_hash(struct net_device *dev, struct sk_buff *skb)
{
	u32 addr1, addr2, ports;
	u32 hash, ihl;
	u8 ip_proto = 0;
	
	if (unlikely(!simple_tx_hashrnd_initialized)) {
		get_random_bytes(&simple_tx_hashrnd, 4);
		simple_tx_hashrnd_initialized = 1;
	}
	
	switch (skb->protocol) {
		case htons(ETH_P_IP):
			if (!(ip_hdr(skb)->frag_off & htons(IP_MF | IP_OFFSET)))
				ip_proto = ip_hdr(skb)->protocol;
			addr1 = ip_hdr(skb)->saddr;
			addr2 = ip_hdr(skb)->daddr;
			ihl = ip_hdr(skb)->ihl;
			break;
		case htons(ETH_P_IPV6):
			ip_proto = ipv6_hdr(skb)->nexthdr;
			addr1 = ipv6_hdr(skb)->saddr.s6_addr32[3];
			addr2 = ipv6_hdr(skb)->daddr.s6_addr32[3];
			ihl = (40 >> 2);
			break;
		default:
			return 0;
	}
	
	
	switch (ip_proto) {
		case IPPROTO_TCP:
		case IPPROTO_UDP:
		case IPPROTO_DCCP:
		case IPPROTO_ESP:
		case IPPROTO_AH:
		case IPPROTO_SCTP:
		case IPPROTO_UDPLITE:
			ports = *((u32 *) (skb_network_header(skb) + (ihl * 4)));
			break;
			
		default:
			ports = 0;
			break;
	}
	
	hash = jhash_3words(addr1, addr2, ports, simple_tx_hashrnd);
	
	return (u16) (((u64) hash * dev->real_num_tx_queues) >> 32);
}

struct dev_select_queue
{
	int netslice_select_queue;
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 28)
	const struct net_device_ops *ndo;
#else
	u16 (*select_queue)(struct net_device *, struct sk_buff *);
#endif
};
static struct dev_select_queue *devs_select_queue = NULL;
static int devs_select_queues = 0;

/* this will be called instead of the original dev_select_queue */
static u16 new_dev_select_queue(struct net_device *dev, struct sk_buff *skb)
{
	if (skb->destructor && skb->destructor == netslice_skb_destructor)
	{
		struct netslice *net_slice = NETSLICE_CB(skb)->net_slice;
		net_slice->stats.tx_select_q ++;
		return (u16)(net_slice->cpu & (dev->real_num_tx_queues - 1));
//		return (u16)(net_slice->cpu % (dev->real_num_tx_queues - 1));		// Ki Suh
	}
	else
	{
		if (dev->ifindex < devs_select_queues)
		{
			struct dev_select_queue *dsq = &devs_select_queue[dev->ifindex];
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 28)
			if (dsq->netslice_select_queue && dsq->ndo)
				return dsq->ndo->ndo_select_queue(dev, skb);
#else
			if (dsq->netslice_select_queue && dsq->select_queue)
				return dsq->select_queue(dev, skb);
#endif
		}
		return (dev->real_num_tx_queues > 1) ? simple_tx_hash(dev, skb) : 0;
	}
}
#endif
#endif //TX_SELECT_QUEUE

static ssize_t netslice_writev(struct file *file, const char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	struct skb_list_head *snd_queue = &net_slice->snd_queue;
	struct netslice_iov iov_fast[UIO_FASTIOV];
	struct netslice_iov *iov = iov_fast, *free_iov = NULL;
	struct netslice_iov __user *iov_user = (struct netslice_iov __user *) buf;
	ssize_t retval = 0, total_bytes;
	int seg, snd_segs;
	
	DECLARE_WAITQUEUE(wait, current);
	
	if (unlikely(count == 0))
	{
		retval = 0;
		goto out;
	}
	
	if (unlikely(count > UIO_MAXIOV))
	{
		retval = -EINVAL;
		goto out;
	}
	
	if (count > UIO_FASTIOV)
	{
		iov = kmalloc(count * sizeof(*iov), GFP_KERNEL);
		if (iov == NULL)
		{
			retval = -ENOMEM;
			goto out;
		}
		free_iov = iov;
	}
	
	/* I will be writing back the iov_rlen */
	if (unlikely(!access_ok(VERIFY_WRITE, iov_user, count * sizeof(*iov_user))))
	{
		retval = -EFAULT;
		goto out;
	}
	
	if (unlikely(__copy_from_user(iov, iov_user, count * sizeof(*iov_user))))
	{
		retval = -EFAULT;
		goto out;
	}
	
	/* check parameter validity */
	total_bytes = 0;
	for (seg = 0; seg < count; seg++)
	{
		void __user *ubuf = iov[seg].iov_base;
		ssize_t len = (ssize_t) iov[seg].iov_rlen;
			
		/* invalid or overflow */
		if (unlikely(len < 0 || (total_bytes + len < total_bytes)))
		{
			retval = -EINVAL;
			goto out;
		}
		
		if (unlikely(!access_ok(VERIFY_READ, ubuf, len)))
		{
			retval = -EINVAL;
			goto out;
		}
		
		total_bytes += len;
	}
	
	
	add_wait_queue(&net_slice->wait, &wait);
	
	for (;;)
	{
		__set_current_state(TASK_INTERRUPTIBLE);

		/* trying to fit the most bytes (i.e. all packets) first
		the base unit of atomic send is the packet */
		for (snd_segs = count; snd_segs > 0; snd_segs--)
		{
			if (atomic_add_return(total_bytes, &snd_queue->pending_bytes_len) < 
				snd_queue->capacity)
				break;
			else
				total_bytes = -((ssize_t)iov[snd_segs-1].iov_rlen);
		}
		if (snd_segs > 0) /* there are snd_segs to send */
			break;
		
		if (signal_pending(current))
		{
			retval = -EINTR;
			break;
		}
		else if (file->f_flags & O_NONBLOCK)
		{
			retval = -EAGAIN;
			break;
		}
		schedule(); /* perhaps I should use schedule_timeout() */
	}
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&net_slice->wait, &wait);
	
	if (unlikely(retval < 0))
		goto out;
	
	if (net_slice->stats.max_snd_burst < snd_segs)
		net_slice->stats.max_snd_burst = snd_segs;
	
	for (seg = 0; seg < snd_segs; seg ++)
	{
		void __user *ubuf;
		size_t len;
		struct sk_buff *skb;
		
		/* packet marked to be skipped by the kernel */
		if (iov[seg].flags & NETSLICE_IOV_SKIP_PACKET_MASK)
			continue;
		
		ubuf = iov[seg].iov_base;
		len = iov[seg].iov_rlen;
		skb = alloc_skb(LL_MAX_HEADER + len, GFP_ATOMIC);
		if (unlikely(skb == NULL))
		{
			retval = -ENOMEM;
			printk(KLOG ": Unable to allocate skb\n");
			goto out;
		}
		skb->ip_summed = CHECKSUM_NONE;
		skb_reserve(skb, LL_MAX_HEADER);
		skb_reset_network_header(skb);
		skb_put(skb, len);
		
		if (unlikely(__copy_from_user(skb->data, ubuf, len) != 0))
		{
			kfree_skb(skb);
			retval = -EFAULT;
			goto out;
		}
		
		NETSLICE_CB(skb)->net_slice = net_slice;
		NETSLICE_CB(skb)->len = skb->len;
		skb->destructor = netslice_skb_destructor;
		do_route_me_harder(&skb);
		
		/* re-compute IP and TCP csums in software, if needed */
		if (iov[seg].flags & NETSLICE_IOV_CSUM_MASK)
		{
			if (iov[seg].flags & NETSLICE_IOV_CSUM_TRANSPORT_MASK)
				transport_ip_checksum_skb(skb);
			else
				ip_checksum_skb(skb);
		}
		else
		{
			net_slice->pre_tx_csum(skb);
		}
		transmit_skb(skb); /* dev_queue_xmit(skb) needs LL header */
	}
	retval = snd_segs;
	
	out:
	if (free_iov)
		kfree(free_iov);
	return retval;
}

static ssize_t netslice_write_default(struct file *file, const char __user *buf, 
	size_t count, loff_t * ppos)
{
	struct netslice *net_slice = file->private_data;
	struct skb_list_head *snd_queue = &net_slice->snd_queue;
	struct sk_buff *skb;
	ssize_t retval = 0;
	
	DECLARE_WAITQUEUE(wait, current);
	
	if (unlikely(count == 0))
	{
		retval = 0;
		goto out;
	}
	
	if (unlikely(!access_ok(VERIFY_READ, buf, count)))
	{
		retval = -EINVAL;
		goto out;
	}
	
	add_wait_queue(&net_slice->wait, &wait);
	
	for (;;)
	{
		__set_current_state(TASK_INTERRUPTIBLE);
	
		if (atomic_add_return(count, &snd_queue->pending_bytes_len) < 
				snd_queue->capacity)
			break;
		
		/* otherwise revert back the value, atomically */
		atomic_sub(count, &snd_queue->pending_bytes_len);
		if (signal_pending(current))
		{
			retval = -EINTR;
			break;
		}
		else if (file->f_flags & O_NONBLOCK)
		{
			retval = -EAGAIN;
			break;
		}
		schedule(); /* perhaps I should use schedule_timeout() */
	}
	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&net_slice->wait, &wait);
	
	if (unlikely(retval < 0))
		goto out;
	
	skb = alloc_skb(LL_MAX_HEADER + count, GFP_ATOMIC);
	if (unlikely(skb == NULL))
	{
		retval = -ENOMEM;
		printk(KLOG ": Unable to allocate skb\n");
		goto out;
	}
	skb->ip_summed = CHECKSUM_NONE;
	skb_reserve(skb, LL_MAX_HEADER);
	skb_reset_network_header(skb);
	skb_put(skb, count);
	if (unlikely(__copy_from_user(skb->data, buf, count) != 0))
	{
		kfree_skb(skb);
		retval = -EFAULT;
		goto out;
	}
	retval = skb->len;
	
	NETSLICE_CB(skb)->net_slice = net_slice;
	NETSLICE_CB(skb)->len = skb->len;
	/* printk(KLOG ": Original skb destructor %p\n", skb->destructor); */
	skb->destructor = netslice_skb_destructor;
	do_route_me_harder(&skb);
	/* re-compute IP and TCP csums in software, if needed */
	net_slice->pre_tx_csum(skb);
	
	transmit_skb(skb); /* dev_queue_xmit(skb) needs LL header */
	
	out:
	return retval;
}

static unsigned int netslice_poll(struct file * file, poll_table *wait)
{
	struct netslice *net_slice = file->private_data;
	int mask = 0;
	
	poll_wait(file, &net_slice->wait, wait);
	
	/* receive queue */
	spin_lock_bh(&net_slice->rcv_queue.lock);
	if (!list_empty(&net_slice->rcv_queue.skbs))
	{
		mask |= (POLLIN | POLLRDNORM);
	}
	spin_unlock_bh(&net_slice->rcv_queue.lock);
	
	/* snd queue, at least 1 byte available */
	if (atomic_read(&net_slice->snd_queue.pending_bytes_len) + SKB_SKACK_MTU < 
		net_slice->snd_queue.capacity)
	{
		mask |= (POLLOUT | POLLWRNORM);
	}	
	return mask;
}

static ssize_t netslice_aio_read(struct kiocb *iocb, const struct iovec *iov, 
	unsigned long nr_segs, loff_t pos)
{
	return 0;
}

static ssize_t netslice_aio_write(struct kiocb *iocb, const struct iovec *iov, 
	unsigned long nr_segs, loff_t pos)
{
	return 0;
}

/* e.g. for ip fragments that are modified in user-space */
static void ip_checksum_skb(struct sk_buff *skb)
{
	struct iphdr *iph = skb_ip_hdr(skb);
	iph->check = 0;
	iph->check = ip_fast_csum(skb_network_header(skb), iph->ihl);
	skb->ip_summed = CHECKSUM_NONE; /* computed the checksum ourselves */
}

static void transport_ip_checksum_skb(struct sk_buff *skb)
{
	struct iphdr *iph = skb_ip_hdr(skb);
	if (iph->protocol == IPPROTO_TCP)
	{
		int iphdr_len = iph->ihl * 4;
		struct tcphdr *th = (struct tcphdr *)((void *)iph + iphdr_len);
		int tcp_pack_len = skb->len - iphdr_len;
		
		th->check = 0;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
		th->check = tcp_v4_check(th, tcp_pack_len, iph->saddr, iph->daddr,
				csum_partial((char *) th, tcp_pack_len, 0));
#else
		th->check = tcp_v4_check(tcp_pack_len, iph->saddr, iph->daddr,
				csum_partial((char *)th, tcp_pack_len, 0));
#endif
	}
	ip_checksum_skb(skb);
}

#define PORTS_RANGE(sport, dport, filter)	( \
		(filter->sport_cnt <= 0 || (sport >= filter->sport_base && \
			sport < filter->sport_base + filter->sport_cnt)) && \
		(filter->dport_cnt <= 0 || (dport >= filter->dport_base && \
			dport < filter->dport_base + filter->dport_cnt)) )

static int default_ip_match(struct netslice *net_slice, struct iphdr *iph)
{
	int is_ip_fragment = 0;
	struct netslice_filter *filter = &net_slice->filter;
	
	if (unlikely(iph->frag_off & ntohs(IP_OFFSET)))
	{
		is_ip_fragment = 1;
		net_slice->stats.ipfrags++; /* count fragments even if dropped (debug) */
	}
	
	if (!( (filter->saddr & filter->smask) == (iph->saddr & filter->smask) ))
		return 0;

	if (!( (filter->daddr & filter->dmask) == (iph->daddr & filter->dmask) ))
		return 0;
	
	if (iph->protocol == IPPROTO_UDP)
	{
		struct udphdr *udph = ip_udp_hdr(iph);
		
		if (!(filter->ipproto & NETSLICE_FILTER_IPPROTO_UDP))
			return 0;
		
		if (is_ip_fragment)
			return (filter->ipproto & NETSLICE_FILTER_IPPROTO_FRAG);
		else
			return PORTS_RANGE(ntohs(udph->source), ntohs(udph->dest), filter);
	}
	else if (iph->protocol == IPPROTO_TCP)
	{
		struct tcphdr *tcph = ip_tcp_hdr(iph);
		
		if (!(filter->ipproto & NETSLICE_FILTER_IPPROTO_TCP))
			return 0;
		
		if (is_ip_fragment)
			return (filter->ipproto & NETSLICE_FILTER_IPPROTO_FRAG);
		else
			return PORTS_RANGE(ntohs(tcph->source), ntohs(tcph->dest), filter);
	}
	else
		return 0;	
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
static unsigned int netslice_nf_hook(unsigned int hook, struct sk_buff** pskb,
		const struct net_device *indev, const struct net_device *outdev,
		int(*okfn)(struct sk_buff*))
{
	struct sk_buff *skb = *pskb; /* peel off one layer of indirection */
#else
static unsigned int netslice_nf_hook(unsigned int hook, struct sk_buff* skb,
		const struct net_device *indev, const struct net_device *outdev,
		int (*okfn) (struct sk_buff*))
{
#endif
		
	/* netfilter hook on PF_INET, hence all traffic is IP */
	struct iphdr *iph = skb_ip_hdr(skb);
	/* don't need to disable preemption here due to softirq context */
	struct netslice *net_slice = &__get_cpu_var(per_cpu_netslice);
	
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 24)
	if (hook == NF_IP_FORWARD)
#else
	if (hook == NF_INET_FORWARD)
#endif
		BUG_ON(indev == NULL); /* on the forwarding path, hence indev != NULL */
	
	/* be polite and don't interfere on the no_iface interface */
	if ((netslice_res.no_iface) && 
		((indev != NULL && netslice_res.no_iface_ifindex == indev->ifindex) || 
		(outdev != NULL && netslice_res.no_iface_ifindex == outdev->ifindex) ))
	{
		net_slice->stats.noiface++;
		goto out_fail;
	}
	
	if (likely(net_slice->ip_match(net_slice, iph)))
	{
		/*
		 * - get the per-CPU apropriate netslice_stats (corresponds to cdev minor)
		 *		- need not disable preemption, since context is softirq
		 * - enqueue the packet (sk_buff)
		 *		- lock, check for space, enqueue, update stats, unlock wakeup
		 */
		int err = 0;
		struct skb_list *skb_list_elem =
#ifdef USE_KMEM_CACHE_WORK
			kmem_cache_alloc(skb_list_cache, GFP_ATOMIC);
#else
			kmalloc(sizeof(*skb_list_elem), GFP_ATOMIC);		
#endif // USE_KMEM_CACHE_WORK
		
		if (unlikely(!skb_list_elem))
		{
			printk(KLOG ": Unable to allocate struct dedup_skb_work\n");
			goto out_fail;
		}
		
		/* linearize packet */
		if (unlikely(skb_is_nonlinear(skb)))
		{
			net_slice->stats.skb_nonlinear++;
			if ( (err = skb_linearize(skb)) != 0)
			{
				printk(KLOG ": Unable to linearize skb\n");
				kfree(skb_list_elem);
				return NF_DROP;
			}
		}
		
		skb_list_elem->skb = skb;

#ifdef USE_SKB_LIST_CPU_INFO_DEBUG
		skb_list_elem->cpu = smp_processor_id();
#endif /*USE_SKB_LIST_CPU_INFO_DEBUG*/

		spin_lock(&net_slice->rcv_queue.lock);
		
		if (net_slice->rcv_queue.bytes_len + skb->len > 
			net_slice->rcv_queue.capacity)
		{
			net_slice->rcv_queue.dropped++;
			err = -ENOMEM;
		}
		else
		{
			list_add_tail(&skb_list_elem->list, &net_slice->rcv_queue.skbs);
			net_slice->rcv_queue.bytes_len += skb->len;
			net_slice->rcv_queue.len++;
		}
		spin_unlock(&net_slice->rcv_queue.lock);
		
		/*
		printk(KLOG ": skb->ip_summed: %d, skb->csum: %x\n", 
			skb->ip_summed, skb->csum); */
		
		/*
		print_symbol(KLOG ": Original skb destructor: %s\n", 
			(unsigned long) skb->destructor); */
		
		/*
		printk(KLOG ": From "NIPQUAD_FMT" to "NIPQUAD_FMT"\n", 
			NIPQUAD(iph->saddr), NIPQUAD(iph->daddr)); */
		
		if (unlikely(err))
		{
			net_slice->stats.nf_hook_dropped++;
			kfree(skb_list_elem);
			return NF_DROP;
		}
		
		net_slice->stats.cnt++;
		/* rise and shine, well, more like rise and read the available packet */
		wake_up_interruptible(&net_slice->wait); /* may also use wake_up() */
		return NF_STOLEN;
	}
	out_fail: //
	net_slice->stats.nomatch++;
	return NF_ACCEPT;
}

/* procfile function called to output data to userspace through /proc */
int proc_put_userland(char *page, char **start, off_t off, int count, 
		int *eof, void *data)
{
	int len = 0, cpu;
	char *dest = (char *) page;
	
	for_each_online_cpu(cpu)
	{
		struct netslice *net_slice = &per_cpu(per_cpu_netslice, cpu);
		struct netslice_stats *s = &net_slice->stats;
		
		len += sprintf(dest, 
			"CPU %d:\n"
				"\tcnt: %ld, no_if: %d, ipfrag: %d, nomatch: %d, dropped: %d\n"
				"\trcv_q: %ld, snd_q: %ld, skb_nonlinear: %d, tx_select_q: %d\n"
				"\tpre_tx_csum (0=none, 1=IP, 2=Transport): %d\n"
				"\treadv: %d, max_r_burst: %d, rcv_trunc: %d, "
				"writev: %d, max_w_burst: %d\n",
			cpu, 
			s->cnt, s->noiface, s->ipfrags, s->nomatch, s->nf_hook_dropped, 
			net_slice->rcv_queue.bytes_len, net_slice->snd_queue.bytes_len, 
			s->skb_nonlinear, s->tx_select_q,
			get_pre_tx_csum(net_slice), 
			net_slice->read == netslice_read_default ? 0 : 1,
			net_slice->stats.max_rcv_burst, net_slice->stats.rcv_trunc,
			net_slice->write == netslice_write_default ? 0 : 1,
			net_slice->stats.max_snd_burst);
		dest = &(((char *) page)[len]);
	}
	
	PROC_READ_RETURN(page, start, off, count, eof, len);
}


/* procfile read data from userspace */
int proc_get_userland(struct file *filp, const char __user *buff,
		unsigned long count, void *data)
{
	if (!procfile_page)
		return count;

	if (count> PAGE_SIZE)
		return -EINVAL;

	if (copy_from_user(procfile_page, buff, count))
		return -EFAULT;

	if (strncmp(procfile_page, 
		COUNTERS_RESET_STRING, strlen(COUNTERS_RESET_STRING)) == 0)
	{
		int cpu;
		for_each_online_cpu(cpu) /* what about cpus going offline? darn */
		{
			struct netslice *net_slice = &per_cpu(per_cpu_netslice, cpu);
			/* safe since we don't depend on the stats to be consistent */
			memset(&net_slice->stats, 0, sizeof(net_slice->stats));
		}
	}

	return count;
}

/* module initialization - called at module load time */
static int __init netslice_init(void)
{
	int i, err = 0, cpu, max_if_idx = 0;
	struct net_device *iface_dev;
#ifdef TX_SELECT_QUEUE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	struct dev_select_queue *dev_sq;
#endif
#endif //TX_SELECT_QUEUE
	struct netslice *net_slice;
	
	cdevs_minor = num_online_cpus();
	
	if (wmem_max < MIN_WMEM_MAX)
	{
		printk(KLOG ": Correcting too small wmem_max parameter from %d to %d\n", 
			wmem_max, MIN_WMEM_MAX);
		wmem_max = MIN_WMEM_MAX;
	}
	if (rmem_max < MIN_RMEM_MAX)
	{
		printk(KLOG ": Correcting too small rmem_max parameter from %d to %d\n", 
			rmem_max, MIN_RMEM_MAX);
		rmem_max = MIN_RMEM_MAX;
	}
	
	for_each_online_cpu(cpu)
	{
		net_slice = &per_cpu(per_cpu_netslice, cpu);
		
		memset(net_slice, 0, sizeof(*net_slice));
		
		init_waitqueue_head(&net_slice->wait);
		
		INIT_LIST_HEAD(&net_slice->rcv_queue.skbs);
		spin_lock_init(&net_slice->rcv_queue.lock);
		atomic_set(&net_slice->rcv_queue.pending_bytes_len, 0);
		net_slice->rcv_queue.capacity = rmem_max;

		INIT_LIST_HEAD(&net_slice->snd_queue.skbs);
		spin_lock_init(&net_slice->snd_queue.lock);
		atomic_set(&net_slice->snd_queue.pending_bytes_len, 0);
		net_slice->snd_queue.capacity = wmem_max;
		
		net_slice->read = netslice_read_default;
		net_slice->write = netslice_write_default;
		net_slice->pre_tx_csum = pre_tx_csum_fn[NETSLICE_PRE_TX_CSUM_NONE];
		net_slice->cpu = cpu;
		net_slice->ip_match = default_ip_match;
		atomic_set(&net_slice->refcnt, 0);
	}
		
	memset(&netslice_res, 0, sizeof(netslice_res));
	
	/* get no_iface net_device index for fast(er) int cmp instead of strcmp */
	if ((strcmp(no_iface, "") != 0)
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 20)
			&& ((iface_dev = dev_get_by_name(no_iface)) != NULL))
#else
			&& ((iface_dev = dev_get_by_name(&init_net, no_iface)) != NULL))
#endif		
	{
		netslice_res.no_iface = 1;
		netslice_res.no_iface_ifindex = iface_dev->ifindex;
		dev_put(iface_dev);
	}

#ifdef TX_SELECT_QUEUE
	/* all devices except for the no_iface */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	read_lock(&dev_base_lock); /* rtnl_lock(); */
	for_each_netdev(&init_net, iface_dev)
	{
		dev_hold(iface_dev);
		if (iface_dev->ifindex > max_if_idx)
			max_if_idx = iface_dev->ifindex;
		dev_put(iface_dev);
	}
	devs_select_queues = max_if_idx+1;
	if (!(devs_select_queue = kzalloc(sizeof(struct dev_select_queue) * 
			devs_select_queues, GFP_ATOMIC)))
	{
		printk(KLOG ": Unable to allocate devs_select_queue\n");
		read_unlock(&dev_base_lock);
		goto out_free_netdev_select_queue;
	}
	for_each_netdev(&init_net, iface_dev)
	{
		dev_hold(iface_dev);
		if (strncmp(iface_dev->name, no_iface, IFNAMSIZ) != 0)
		{
			dev_sq = &devs_select_queue[iface_dev->ifindex];
			BUG_ON(dev_sq->netslice_select_queue); /* it's one shot */
			dev_sq->netslice_select_queue = 1;
			/* save old and plug in the new function */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 28)
			dev_sq->select_queue = iface_dev->select_queue;
			iface_dev->select_queue = new_dev_select_queue;
#else
			dev_sq->ndo = iface_dev->netdev_ops;
			{
				struct net_device_ops *ndo = kmalloc(sizeof(*ndo), GFP_KERNEL);
				if (!ndo)
				{
					printk(KLOG ": Unable to allocate struct net_device_ops\n");
					dev_put(iface_dev);
					goto out_free_netdev_select_queue;
				}
				
				memcpy(ndo, iface_dev->netdev_ops, sizeof(*ndo));
				ndo->ndo_select_queue = new_dev_select_queue;
				iface_dev->netdev_ops = ndo;
			}
#endif
		}
		dev_put(iface_dev);
	}
	read_unlock(&dev_base_lock); /* rtnl_unlock(); */
#endif
#endif //TX_SELECT_QUEUE
	
#ifdef USE_KMEM_CACHE_SKB_LIST
	if (!(skb_list_cache = KMEM_CACHE(skb_list, SLAB_HWCACHE_ALIGN|SLAB_PANIC)))
	{
		printk(KLOG ": Unable to allocate kmem-cache\n");
		err = -ENOMEM;
		goto out_no_kmem_cache;
	}
#endif /* USE_KMEM_CACHE_SKB_LIST */
	
	if (!(procfile_page = (void *) get_zeroed_page(GFP_KERNEL)))
	{
		printk(KLOG ": Unable to allocate procfile page\n");
		err = -ENOMEM;
		goto out;
	}
	/* create the proc file */
	if (!(proc_entry = create_proc_entry(procfile, 0666, NULL)))
	{
		printk(KLOG ": Unable to create proc file %s\n", procfile);
		err = -EINVAL;
		goto out_free_procfile_page;
	}
	proc_entry->write_proc = proc_get_userland;
	proc_entry->read_proc = proc_put_userland;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
	proc_entry->owner = THIS_MODULE;
#endif
	
	err = alloc_chrdev_region(&netslice_dev, 0, cdevs_minor, CHR_DEV_NAME);
	if (err < 0)
	{
		printk(KLOG ": Unable to allocate major number for char device\n");
		goto out_free_procfile;
	}
	
	cdev_init(&netslice_cdev, &netslice_fops);
	err = cdev_add(&netslice_cdev, netslice_dev, cdevs_minor);
	if (err < 0)
	{
		printk(KLOG ": Unable to allocate chrdev\n");
		goto out_free_chrdev;
	}
	
	/* netfilter hook */
	memset(&netslice_nf_hook_ops, 0, sizeof(netslice_nf_hook_ops));
	netslice_nf_hook_ops.hook = netslice_nf_hook;
	netslice_nf_hook_ops.hooknum = hook_chain;
	netslice_nf_hook_ops.pf = PF_INET;
	netslice_nf_hook_ops.priority = hook_prio;
	err = nf_register_hook(&netslice_nf_hook_ops);
	
	printk(KLOG ": %s loaded\n", CHR_DEV_NAME);
	return err;
	
	out_free_chrdev:
	unregister_chrdev_region(netslice_dev, cdevs_minor);
	
	out_free_procfile:
	if (proc_entry)
		remove_proc_entry(procfile, NULL);
	
	out_free_procfile_page:
	if (procfile_page)
		free_page((unsigned long) procfile_page);
	
	out:
#ifdef USE_KMEM_CACHE_SKB_LIST
	if (skb_list_cache)
		kmem_cache_destroy(skb_list_cache);
#endif
#ifdef TX_SELECT_QUEUE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	out_free_netdev_select_queue:
	
	for (i = 0; i < devs_select_queues; i++)
	{
		dev_sq = &devs_select_queue[i];
		if (!dev_sq->netslice_select_queue)
			continue;
		
		iface_dev = dev_get_by_index(&init_net, i);
		if (iface_dev)
		{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 28)
			if (iface_dev->select_queue && 
					iface_dev->select_queue == new_dev_select_queue)
				iface_dev->select_queue = dev_sq->select_queue;
#else
			if (iface_dev->netdev_ops && 
				iface_dev->netdev_ops->ndo_select_queue == new_dev_select_queue)
			{
				const struct net_device_ops *tmp_ndo = iface_dev->netdev_ops;
				iface_dev->netdev_ops = dev_sq->ndo;
				kfree(tmp_ndo);
			}
#endif
			dev_put(iface_dev);
		}
	}
	kfree(devs_select_queue);
	devs_select_queue = NULL;
#endif
#endif //TX_SELECT_QUEUE
#ifdef USE_KMEM_CACHE_SKB_LIST
	out_no_kmem_cache:
#endif /* USE_KMEM_CACHE_SKB_LIST */
	return err;
}

/* module unload, only succesful after the chrdev file descriptor is released */
static void __exit netslice_exit(void)
{
	int cpu, i;
	DECLARE_WAITQUEUE(wait, current);
	DECLARE_WAIT_QUEUE_HEAD(rmmod_waiters);
#ifdef TX_SELECT_QUEUE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	struct net_device *iface_dev;
	struct dev_select_queue *dev_sq;
#endif
#endif //TX_SELECT_QUEUE
	
	if (proc_entry)
		remove_proc_entry(procfile, NULL);
	if (procfile_page)
		free_page((unsigned long) procfile_page);
	
	if (netslice_nf_hook_ops.hook) /* hook function should have been set */
	{
		nf_unregister_hook(&netslice_nf_hook_ops);
		memset(&netslice_nf_hook_ops, 0, sizeof(netslice_nf_hook_ops));
	}
	
	cdev_del(&netslice_cdev);
	unregister_chrdev_region(netslice_dev, cdevs_minor);
	
	/* clean up the rx queues */
	for_each_online_cpu(cpu)
	{
		struct skb_list *elem, *tmp;
		struct netslice *net_slice = &per_cpu(per_cpu_netslice, cpu);
		
		spin_lock(&net_slice->rcv_queue.lock);
		list_for_each_entry_safe(elem, tmp, &net_slice->rcv_queue.skbs, list)
		{
			if (elem->skb)
				kfree_skb(elem->skb);
			kfree(elem);
		}
		spin_unlock(&net_slice->rcv_queue.lock);
	}
	printk(KLOG ": cleaned up rx queues\n");
	/* wait for all tx queues, tricky since it's callback based */
	add_wait_queue(&rmmod_waiters, &wait);
	for (;;)
	{
		int all_queues_depleted = 1;

		set_current_state(TASK_INTERRUPTIBLE);
		
		for_each_online_cpu(cpu)
		{
			struct netslice *net_slice = &per_cpu(per_cpu_netslice, cpu);
			if (atomic_read(&net_slice->snd_queue.pending_bytes_len) > 0)
			{
				all_queues_depleted = 0;
				break;
			}
		}
		if (all_queues_depleted)
			break;
		
		schedule_timeout(1);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&rmmod_waiters, &wait);
	printk(KLOG ": cleaned up tx queues\n");
	
#ifdef USE_KMEM_CACHE_SKB_LIST
	if (skb_list_cache)
		kmem_cache_destroy(skb_list_cache);
#endif /* USE_KMEM_CACHE_SKB_LIST */

#ifdef TX_SELECT_QUEUE
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	for (i = 0; i < devs_select_queues; i++)
	{
		dev_sq = &devs_select_queue[i];
		if (!dev_sq->netslice_select_queue)
			continue;
		
		iface_dev = dev_get_by_index(&init_net, i);
		if (iface_dev)
		{

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 28)
			if (iface_dev->select_queue && 
					iface_dev->select_queue == new_dev_select_queue)
				iface_dev->select_queue = dev_sq->select_queue;
#else
			if (iface_dev->netdev_ops && 
				iface_dev->netdev_ops->ndo_select_queue == new_dev_select_queue)
			{
				const struct net_device_ops *tmp_ndo = iface_dev->netdev_ops;
				iface_dev->netdev_ops = dev_sq->ndo;
				kfree(tmp_ndo);
			}
#endif
			dev_put(iface_dev);
		}
	}
	kfree(devs_select_queue);
	devs_select_queue = NULL;
#endif
#endif //TX_SELECT_QUEUE
	printk(KLOG ": %s removed\n", CHR_DEV_NAME);
}

module_init(netslice_init);
module_exit(netslice_exit);

MODULE_DESCRIPTION("netslice driver");
MODULE_AUTHOR("Tudor Marian <tudor.marian@cornell.edu>");
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(a)
#endif
MODULE_LICENSE("Dual BSD/GPL");

